import numba_stats as nb
import scipy.stats as sc
import numpy as np
import pytest


@pytest.mark.benchmark(group="norm.pdf")
@pytest.mark.parametrize("which", ("scipy", "ours"))
@pytest.mark.parametrize("n", (10, 100, 1000, 10000))
def test_norm_pdf_speed(benchmark, which, n):
    x = np.linspace(-5, 5, n)
    m = np.linspace(-1, 1, n)
    s = np.linspace(0.1, 1, n)
    benchmark(
        lambda: sc.norm.pdf(x, m, s) if which == "scipy" else nb.norm.pdf(x, m, s)
    )


@pytest.mark.benchmark(group="norm.cdf")
@pytest.mark.parametrize("which", ("scipy", "ours"))
@pytest.mark.parametrize("n", (10, 100, 1000, 10000))
def test_norm_cdf_speed(benchmark, which, n):
    x = np.linspace(-5, 5, n)
    m = np.linspace(-1, 1, n)
    s = np.linspace(0.1, 1, n)
    benchmark(
        lambda: sc.norm.cdf(x, m, s) if which == "scipy" else nb.norm.cdf(x, m, s)
    )


@pytest.mark.benchmark(group="norm.ppf")
@pytest.mark.parametrize("which", ("scipy", "ours"))
@pytest.mark.parametrize("n", (10, 100, 1000, 10000))
def test_norm_ppf_speed(benchmark, which, n):
    x = np.linspace(0, 1, n)
    m = np.linspace(-1, 1, n)
    s = np.linspace(0.1, 1, n)
    benchmark(
        lambda: sc.norm.ppf(x, m, s) if which == "scipy" else nb.norm.ppf(x, m, s)
    )


@pytest.mark.benchmark(group="poisson.pmf")
@pytest.mark.parametrize("which", ("scipy", "ours"))
@pytest.mark.parametrize("n", (10, 100, 1000, 10000))
def test_poisson_pmf_speed(benchmark, which, n):
    k = np.arange(0, n)
    m = np.linspace(0, 10, n)
    benchmark(
        lambda: sc.poisson.pmf(k, m) if which == "scipy" else nb.poisson.pmf(k, m)
    )


@pytest.mark.benchmark(group="poisson.cdf")
@pytest.mark.parametrize("which", ("scipy", "ours"))
@pytest.mark.parametrize("n", (10, 100, 1000, 10000))
def test_poisson_cdf_speed(benchmark, which, n):
    k = np.arange(0, n)
    m = np.linspace(0, 10, n)
    benchmark(
        lambda: sc.poisson.cdf(k, m) if which == "scipy" else nb.poisson.cdf(k, m)
    )


@pytest.mark.benchmark(group="expon.pdf")
@pytest.mark.parametrize("which", ("scipy", "ours"))
@pytest.mark.parametrize("n", (10, 100, 1000, 10000))
def test_expon_pdf_speed(benchmark, which, n):
    x = np.linspace(0, 10, n)
    m = np.linspace(0, 10, n)
    s = np.linspace(1, 10, n)
    benchmark(
        lambda: sc.expon.pdf(x, m, s) if which == "scipy" else nb.expon.pdf(x, m, s)
    )


@pytest.mark.benchmark(group="expon.cdf")
@pytest.mark.parametrize("which", ("scipy", "ours"))
@pytest.mark.parametrize("n", (10, 100, 1000, 10000))
def test_expon_cdf_speed(benchmark, which, n):
    x = np.linspace(0, 10, n)
    m = np.linspace(0, 10, n)
    s = np.linspace(1, 10, n)
    benchmark(
        lambda: sc.expon.cdf(x, m, s) if which == "scipy" else nb.expon.cdf(x, m, s)
    )


@pytest.mark.benchmark(group="t.pdf")
@pytest.mark.parametrize("which", ("scipy", "ours"))
@pytest.mark.parametrize("n", (10, 100, 1000, 10000))
def test_t_pdf_speed(benchmark, which, n):
    x = np.linspace(-5, 5, n)
    df = np.linspace(1, 10, n)
    m = np.linspace(-1, 1, n)
    s = np.linspace(0.1, 1, n)
    benchmark(
        lambda: sc.t.pdf(x, df, m, s) if which == "scipy" else nb.t.pdf(x, df, m, s)
    )


@pytest.mark.benchmark(group="t.cdf")
@pytest.mark.parametrize("which", ("scipy", "ours"))
@pytest.mark.parametrize("n", (10, 100, 1000, 10000))
def test_t_cdf_speed(benchmark, which, n):
    x = np.linspace(-5, 5, n)
    df = np.linspace(1, 10, n)
    m = np.linspace(-1, 1, n)
    s = np.linspace(0.1, 1, n)
    benchmark(
        lambda: sc.t.cdf(x, df, m, s) if which == "scipy" else nb.t.cdf(x, df, m, s)
    )


@pytest.mark.benchmark(group="t.ppf")
@pytest.mark.parametrize("which", ("scipy", "ours"))
@pytest.mark.parametrize("n", (10, 100, 1000, 10000))
def test_t_ppf_speed(benchmark, which, n):
    x = np.linspace(0, 1, n)
    df = np.linspace(1, 10, n)
    m = np.linspace(-1, 1, n)
    s = np.linspace(0.1, 1, n)
    benchmark(
        lambda: sc.t.ppf(x, df, m, s) if which == "scipy" else nb.t.ppf(x, df, m, s)
    )



from numba_stats import bernstein
from scipy.interpolate import BPoly
import pytest
import numpy as np
from scipy.integrate import quad
import numba as nb


@pytest.mark.parametrize(
    "beta", [[1.0], [1.0, 1.0], [1.0, 1.0, 1.0], [1.0, 2.0, 3.0], [1.0, 3.0, 2.0]]
)
def test_bernstein_density(beta):
    x = np.linspace(1, 3)
    got = bernstein.density(x, beta, x[0], x[-1])
    expected = BPoly(np.array(beta)[:, np.newaxis], [x[0], x[-1]])(x)
    np.testing.assert_allclose(got, expected)

    got = bernstein.density(0.5, beta, 0, 1)
    expected = bernstein.density([0.5], beta, 0, 1)
    np.testing.assert_allclose(got, expected)


@pytest.mark.parametrize("beta", [[1], [1, 1], [1, 1, 1]])
def test_bernstein_integral(beta):
    xrange = 1.5, 3.4
    got = bernstein.scaled_cdf(xrange[1], beta, *xrange)
    expected = 1
    np.testing.assert_allclose(got, expected)


@pytest.mark.parametrize("beta", [[1], [1, 1], [1, 1, 1], [1, 2, 3], [1, 3, 2]])
def test_bernstein_scaled_cdf(beta):
    x = np.linspace(0, 1)

    got = bernstein.scaled_cdf(x, beta, x[0], x[-1])
    expected = [
        quad(lambda y: bernstein.density(y, beta, x[0], x[-1]), x[0], xi)[0] for xi in x
    ]
    np.testing.assert_allclose(got, expected)


def test_numba_bernstein_density():
    @nb.njit
    def f():
        return bernstein.density(
            np.array([0.5, 0.6]),
            np.array([1.0, 2.0, 3.0]),
            0.0,
            1.0,
        )

    f()


def test_numba_bernstein_scaled_cdf():
    @nb.njit
    def f():
        return bernstein.scaled_cdf(
            np.array([0.5, 0.6]),
            np.array([1.0, 2.0, 3.0]),
            0.0,
            1.0,
        )

    f()



import numba as nb
import functools
from typing import Callable


def _vectorize(narg, cache=True, **kwargs):
    def outer(func):
        if "cache" not in kwargs:
            kwargs["cache"] = cache
        # if "error_model" not in kwargs:
        #     kwargs["error_model"] = "numpy"

        signatures = [arg(*([arg] * narg)) for arg in (nb.float32, nb.float64)]

        wrapped = nb.vectorize(signatures, **kwargs)(func)
        # this does not help with the docs, unfortunately
        functools.update_wrapper(wrapped, func)

        return wrapped

    return outer


def _jit(*args, **kwargs):
    if "error_model" not in kwargs:
        kwargs["error_model"] = "numpy"

    if len(args) == 1 and isinstance(args[0], Callable):
        return nb.njit(**kwargs)(args[0])

    def outer(func):
        return nb.njit(**kwargs)(func)

    return outer



"""
Poisson distribution.
"""

import numba as nb
import numpy as np
from ._special import gammaincc as _gammaincc
from math import lgamma as _lgamma

_signatures = [
    nb.float32(nb.int32, nb.float32),
    nb.float64(nb.intp, nb.float64),
]


@nb.vectorize(_signatures)
def logpmf(k, mu):
    """
    Return log of probability mass.
    """
    if mu == 0:
        return 0.0 if k == 0 else -np.inf
    return k * np.log(mu) - _lgamma(k + 1.0) - mu


@nb.vectorize(_signatures)
def pmf(k, mu):
    """
    Return probability mass.
    """
    return np.exp(logpmf(k, mu))


@nb.vectorize(_signatures)
def cdf(k, mu):
    """
    Evaluate cumulative distribution function.
    """
    return _gammaincc(k + 1, mu)



"""
Q-Gaussian distribution.

A generalisation (q-analog) of the normal distribution based on Tsallis entropy. It
can be used an alternative model for the normal distribution to check for model bias.

It is equivalent to Student's t distribution and can be computed from the latter via
a change of variables, which is exploited in this implementation.

https://en.wikipedia.org/wiki/Q-Gaussian_distribution
"""

import numpy as np
from math import lgamma as _lgamma
from . import norm as _norm, t as _t
from ._util import _jit, _vectorize


@_jit
def _qexp(x, q):
    if q == 1:
        return np.exp(x)
    alpha = 1.0 - q
    arg = 1.0 + alpha * x
    if arg <= 0:
        return 0
    le = np.log(arg) / alpha
    return np.exp(le)


@_jit
def _compute_cq(q):
    # beta = 1/2 for equivalence with normal distribution for q = 1
    const = np.sqrt(2 * np.pi)
    alpha = 1.0 - q
    if q == 1:
        return const
    if q < 1:
        return (
            2.0
            * const
            / ((3.0 - q) * np.sqrt(alpha))
            * np.exp(_lgamma(1.0 / alpha) - _lgamma(0.5 * (3.0 - q) / alpha))
        )
    if q < 3:
        return (
            const
            / np.sqrt(-alpha)
            * np.exp(_lgamma(-0.5 * (3.0 - q) / alpha) - _lgamma(-1.0 / alpha))
        )
    return np.nan


@_jit
def _df_sigma(q, sigma):
    # https://en.wikipedia.org/wiki/Q-Gaussian_distribution
    # relation to Student's t-distribution

    # 1/(2 sigma^2) = 1 / (3 - q)
    # 2 sigma^2 = 3 - q
    # sigma = sqrt((3 - q)/2)

    df = (3 - q) / (q - 1)
    sigma /= np.sqrt(0.5 * (3 - q))

    return df, sigma


@_vectorize(4)
def pdf(x, q, mu, sigma):
    inv_scale = 1.0 / sigma
    z = (x - mu) * inv_scale
    c_q = _compute_cq(q)
    inv_scale /= c_q
    # beta = 1/2 for equivalence with normal distribution for q = 1
    if q == 1.0:
        return np.exp(-0.5 * z ** 2) * inv_scale
    return _qexp(-0.5 * z ** 2, q) * inv_scale


@_vectorize(4, cache=False)
def cdf(x, q, mu, sigma):
    if q < 1 or q > 3:
        raise ValueError("q < 1 or q > 3 are not supported")

    if q == 1:
        return _norm.cdf(x, mu, sigma)

    df, sigma = _df_sigma(q, sigma)

    return _t.cdf(x, df, mu, sigma)


@_vectorize(4, cache=False)
def ppf(x, q, mu, sigma):
    if q < 1 or q > 3:
        raise ValueError("q < 1 or q > 3 are not supported")

    if q == 1:
        return _norm.ppf(x, mu, sigma)

    df, sigma = _df_sigma(q, sigma)

    return _t.ppf(x, df, mu, sigma)



from numba_stats import qgaussian, t, norm
import numpy as np
import pytest
from scipy.integrate import quad


def q_sigma(nu, sigma):
    # https://en.wikipedia.org/wiki/Q-Gaussian_distribution
    # relation to Student's t-distribution

    # 1 / (2 sigma^2) = 1 / (3 - q)
    # sqrt((3 - q) / 2) = sigma

    q = (nu + 3.0) / (nu + 1.0)
    return q, sigma * np.sqrt(0.5 * (3.0 - q))


def test_qgaussian_pdf_vs_norm():
    x = np.linspace(-5, 5)

    expected = norm.pdf(x, 0.5, 1.2)
    got = qgaussian.pdf(x, 1, 0.5, 1.2)

    np.testing.assert_allclose(got, expected)


def test_qgaussian_cdf_vs_norm():
    x = np.linspace(-5, 5)

    expected = norm.cdf(x, 0.5, 1.2)
    got = qgaussian.cdf(x, 1, 0.5, 1.2)

    np.testing.assert_allclose(got, expected)


@pytest.mark.parametrize("nu", np.arange(1, 11))
def test_qgaussian_pdf_vs_t(nu):
    x = np.linspace(-5, 5)
    q, sigma = q_sigma(nu, 1.2)

    expected = t.pdf(x, nu, 0.1, 1.2)
    got = qgaussian.pdf(x, q, 0.1, sigma)

    np.testing.assert_allclose(got, expected)


@pytest.mark.parametrize("q", (1, 1.1, 1.2, 1.3, 1.4, 1.5, 1.9, 2.0, 2.5, 2.9))
def test_qgaussian_cdf(q):
    x = np.linspace(-5, 5, 10)

    expected = [quad(lambda y: qgaussian.pdf(y, q, 0.1, 1.2), 0, xi)[0] for xi in x]
    got = qgaussian.cdf(x, q, 0.1, 1.2) - qgaussian.cdf(0, q, 0.1, 1.2)

    np.testing.assert_allclose(got, expected)


@pytest.mark.parametrize("q", (1, 1.1, 1.2, 1.3, 1.4, 1.5, 1.9, 2.0, 2.5, 2.9))
def test_qgaussian_ppf(q):
    x = np.linspace(-5, 5, 10)

    def f(x):
        p = qgaussian.cdf(x, q, 2, 3)
        return qgaussian.ppf(p, q, 2, 3)

    expected = x
    got = f(x)

    np.testing.assert_allclose(got, expected)



"""
Exponential distribution.
"""
import numpy as np
from math import expm1 as _expm1, log1p as _log1p
from ._util import _jit, _vectorize


@_jit
def _cdf(z):
    return -_expm1(-z)


@_jit
def _ppf(p):
    return -_log1p(-p)


@_jit
def _logpdf(x, loc, scale):
    z = (x - loc) / scale
    return -z - np.log(scale)


@_vectorize(3)
def logpdf(x, loc, scale):
    """
    Return log of probability density.
    """
    return _logpdf(x, loc, scale)


@_vectorize(3)
def pdf(x, loc, scale):
    """
    Return probability density.
    """
    return np.exp(_logpdf(x, loc, scale))


@_vectorize(3)
def cdf(x, loc, scale):
    """
    Return cumulative probability.
    """
    z = (x - loc) / scale
    return _cdf(z)


@_vectorize(3)
def ppf(p, loc, scale):
    """
    Return quantile for given probability.
    """
    z = _ppf(p)
    x = z * scale + loc
    return x



"""
Generalised Crystal Ball distribution.

The generalised Crystal Ball distribution replaces the lower and upper tail of
an asymmetric normal distribution with power-law tails. Furthermore, the scale
is allowed to vary between the left and the right side of the peak. There is no
discontinuity at the maximum or elsewhere.
"""

from .crystalball import _powerlaw_integral, _normal_integral, _density
from ._util import _vectorize, _jit


@_jit
def _norm_half(beta, m, scale):
    return (_normal_integral(0, beta) + _powerlaw_integral(-beta, beta, m)) * scale


@_vectorize(8)
def pdf(x, beta_left, m_left, scale_left, beta_right, m_right, scale_right, loc):
    """
    Return probability density.
    """
    if x < loc:
        scale = scale_left
        beta = beta_left
        m = m_left
        z = (x - loc) / scale
    else:
        scale = scale_right
        beta = beta_right
        m = m_right
        z = (loc - x) / scale
    dens = _density(z, beta, m)
    return dens / (
        _norm_half(beta_left, m_left, scale_left)
        + _norm_half(beta_right, m_right, scale_right)
    )


@_vectorize(8)
def cdf(x, beta_left, m_left, scale_left, beta_right, m_right, scale_right, loc):
    """
    Return cumulative probability.
    """
    norm = _norm_half(beta_left, m_left, scale_left) + _norm_half(
        beta_right, m_right, scale_right
    )
    z = x - loc
    if z < 0:
        z /= scale_left
    else:
        z /= scale_right
    if z < -beta_left:
        return _powerlaw_integral(z, beta_left, m_left) / norm * scale_left
    if z < 0:
        return (
            (
                _powerlaw_integral(-beta_left, beta_left, m_left)
                + _normal_integral(-beta_left, z)
            )
            * scale_left
            / norm
        )
    if z < beta_right:
        return (
            (
                _powerlaw_integral(-beta_left, beta_left, m_left)
                + _normal_integral(-beta_left, 0)
            )
            * scale_left
            + _normal_integral(0, z) * scale_right
        ) / norm
    return (
        (
            _powerlaw_integral(-beta_left, beta_left, m_left)
            + _normal_integral(-beta_left, 0)
        )
        * scale_left
        + (
            _normal_integral(0, beta_right)
            + _powerlaw_integral(-beta_right, beta_right, m_right)
            - _powerlaw_integral(-z, beta_right, m_right)
        )
        * scale_right
    ) / norm



"""
Truncated normal distribution.
"""

import numpy as np
from .norm import _logpdf as _norm_logpdf, _cdf, _ppf
from ._util import _jit, _vectorize


@_jit
def _logpdf(z, zmin, zmax):
    if z < zmin or z > zmax:
        return -np.inf
    return _norm_logpdf(z) - np.log(_cdf(zmax) - _cdf(zmin))


@_vectorize(5)
def logpdf(x, xmin, xmax, loc, scale):
    """
    Return log of probability density.
    """
    scale_inv = 1 / scale
    z = (x - loc) * scale_inv
    zmin = (xmin - loc) * scale_inv
    zmax = (xmax - loc) * scale_inv
    return _logpdf(z, zmin, zmax) + np.log(scale_inv)


@_vectorize(5)
def pdf(x, xmin, xmax, loc, scale):
    """
    Return probability density.
    """
    scale_inv = 1 / scale
    z = (x - loc) * scale_inv
    zmin = (xmin - loc) * scale_inv
    zmax = (xmax - loc) * scale_inv
    return np.exp(_logpdf(z, zmin, zmax)) * scale_inv


@_vectorize(5)
def cdf(x, xmin, xmax, loc, scale):
    """
    Return cumulative probability.
    """
    if x < xmin:
        return 0.0
    elif x > xmax:
        return 1.0
    scale_inv = 1 / scale
    z = (x - loc) * scale_inv
    zmin = (xmin - loc) * scale_inv
    zmax = (xmax - loc) * scale_inv
    pmin = _cdf(zmin)
    pmax = _cdf(zmax)
    return (_cdf(z) - pmin) / (pmax - pmin)


@_vectorize(5, cache=False)
def ppf(p, xmin, xmax, loc, scale):
    """
    Return quantile for given probability.
    """
    scale_inv = 1 / scale
    zmin = (xmin - loc) * scale_inv
    zmax = (xmax - loc) * scale_inv
    pmin = _cdf(zmin)
    pmax = _cdf(zmax)
    pstar = p * (pmax - pmin) + pmin
    z = _ppf(pstar)
    return scale * z + loc



"""
Normal distribution.
"""
import numpy as np
from ._special import erfinv as _erfinv
from ._util import _vectorize, _jit
from math import erf as _erf


@_jit
def _logpdf(z):
    return -0.5 * (z ** 2 + np.log(2 * np.pi))


@_jit
def _cdf(z):
    c = np.sqrt(0.5)
    return 0.5 * (1.0 + _erf(z * c))


@_jit(cache=False)  # cannot cache because of _erfinv
def _ppf(p):
    return np.sqrt(2) * _erfinv(2 * p - 1)


@_vectorize(3)
def logpdf(x, loc, scale):
    """
    Return log of probability density.
    """
    z = (x - loc) / scale
    return _logpdf(z) - np.log(scale)


@_vectorize(3)
def pdf(x, loc, scale):
    """
    Return probability density.
    """
    return np.exp(logpdf(x, loc, scale))


@_vectorize(3)
def cdf(x, loc, scale):
    """
    Return cumulative probability.
    """
    z = (x - loc) / scale
    return _cdf(z)


@_vectorize(3, cache=False)  # cannot cache this
def ppf(p, loc, scale):
    """
    Return quantile for given probability.
    """
    z = _ppf(p)
    return scale * z + loc



import scipy.stats as sc
from scipy.integrate import quad
import numpy as np
import numba as nb


def test_njit_with_numba_stats():
    from numba_stats import norm

    @nb.njit
    def test(x):
        p = norm.cdf(x, 0, 1)
        return norm.ppf(p, 0, 1)

    expected = np.linspace(-3, 3, 10)
    got = test(expected)
    np.testing.assert_allclose(got, expected)


def test_uniform_pdf():
    from numba_stats import uniform

    x = np.linspace(-1.1, 1.1, 10)
    got = uniform.pdf(x, -1, 2)
    expected = sc.uniform.pdf(x, -1, 2)
    np.testing.assert_allclose(got, expected)


def test_uniform_cdf():
    from numba_stats import uniform

    x = np.linspace(-1.1, 1.1, 10)
    got = uniform.cdf(x, -1, 2)
    expected = sc.uniform.cdf(x, -1, 2)
    np.testing.assert_allclose(got, expected)


def test_uniform_ppf():
    from numba_stats import uniform

    x = np.linspace(0, 1, 10)
    got = uniform.ppf(x, -1, 2)
    expected = sc.uniform.ppf(x, -1, 2)
    np.testing.assert_allclose(got, expected)


def test_tsallis_pdf():
    from numba_stats import tsallis

    for m in (100, 1000):
        for t in (100, 1000):
            for n in (3, 5, 8):
                v, err = quad(lambda pt: tsallis.pdf(pt, m, t, n), 0, np.inf)
                assert abs(1 - v) < err


def test_tsallis_cdf():
    from numba_stats import tsallis

    for m in (100, 1000):
        for t in (100, 1000):
            for n in (3, 5, 8):
                for ptrange in ((0, 500), (500, 1000), (1000, 2000)):
                    v, err = quad(lambda pt: tsallis.pdf(pt, m, t, n), *ptrange)
                    v2 = np.diff(tsallis.cdf(ptrange, m, t, n))
                    assert abs(v2 - v) < err



"""
Crystal Ball distribution.

The Crystal Ball distribution replaces the lower tail of a normal distribution with
a power-law tail.

https://en.wikipedia.org/wiki/Crystal_Ball_function
"""

from ._util import _jit, _vectorize
import numpy as np
from math import erf as _erf


@_jit
def _powerlaw(z, beta, m):
    assert beta > 0
    assert m > 0
    exp_beta = np.exp(-0.5 * beta ** 2)
    a = (m / beta) ** m * exp_beta
    b = m / beta - beta
    return a * (b - z) ** -m


@_jit
def _powerlaw_integral(z, beta, m):
    assert beta > 0
    assert m > 1
    exp_beta = np.exp(-0.5 * beta ** 2)
    a = (m / beta) ** m * exp_beta
    b = m / beta - beta
    m1 = m - 1
    return a * (b - z) ** -m1 / m1


@_jit
def _normal_integral(a, b):
    sqrt_half = np.sqrt(0.5)
    return sqrt_half * np.sqrt(np.pi) * (_erf(b * sqrt_half) - _erf(a * sqrt_half))


@_jit
def _density(z, beta, m):
    if z <= -beta:
        return _powerlaw(z, beta, m)
    return np.exp(-0.5 * z ** 2)


@_vectorize(5)
def pdf(x, beta, m, loc, scale):
    """
    Return probability density.
    """
    z = (x - loc) / scale
    dens = _density(z, beta, m)
    norm = scale * (
        _powerlaw_integral(-beta, beta, m) + _normal_integral(-beta, np.inf)
    )
    return dens / norm


@_vectorize(5)
def cdf(x, beta, m, loc, scale):
    """
    Return cumulative probability.
    """
    z = (x - loc) / scale
    norm = _powerlaw_integral(-beta, beta, m) + _normal_integral(-beta, np.inf)
    if z <= -beta:
        return _powerlaw_integral(z, beta, m) / norm
    return (_powerlaw_integral(-beta, beta, m) + _normal_integral(-beta, z)) / norm



import scipy.stats as sc
import numpy as np
from numba_stats import expon


def test_pdf():
    x = np.linspace(1, 5, 20)
    got = expon.pdf(x, 1, 2)
    expected = sc.expon.pdf(x, 1, 2)
    np.testing.assert_allclose(got, expected)


def test_cdf():
    x = np.linspace(1, 5, 20) + 3
    got = expon.cdf(x, 3, 2)
    expected = sc.expon.cdf(x, 3, 2)
    np.testing.assert_allclose(got, expected)


def test_ppf():
    p = np.linspace(0, 1, 20)
    with np.errstate(invalid="ignore", divide="ignore"):
        got = expon.ppf(p, 1, 2)
    expected = sc.expon.ppf(p, 1, 2)
    np.testing.assert_allclose(got, expected)



"""
Continuous Poisson distribution.
"""
from ._special import gammaincc as _gammaincc
from ._util import _vectorize


@_vectorize(2, cache=False)
def cdf(x, mu):
    """
    Return cumulative probability.
    """
    return _gammaincc(x + 1, mu)


# The pdf, d cdf(x, mu)/ dx, cannot be expressed in tabulated functions:
#
# d G(x, mu)/d x = ln(mu) G(x, mu) + mu T(3, x, mu)
#
# where G(x, mu) is the upper incomplete gamma function and T(m, s, x) is a special case
# of the Meijer G-function,
# see https://en.wikipedia.org/wiki/Incomplete_gamma_function#Derivatives
#
# There is a Meijer G-function implemented in mpmath, but I don't know how to use it.



import scipy.stats as sc
import numpy as np


def test_lognorm_pdf():
    from numba_stats import lognorm

    x = np.linspace(0, 5, 10)
    got = lognorm.pdf(x, 1.5, 0.1, 1.2)
    expected = sc.lognorm.pdf(x, 1.5, 0.1, 1.2)
    np.testing.assert_allclose(got, expected)


def test_lognorm_logpdf():
    from numba_stats import lognorm

    x = np.linspace(0, 5, 10)
    got = lognorm.logpdf(x, 1.5, 0.1, 1.2)
    expected = sc.lognorm.logpdf(x, 1.5, 0.1, 1.2)
    np.testing.assert_allclose(got, expected)


def test_lognorm_cdf():
    from numba_stats import lognorm

    x = np.linspace(0, 5, 10)
    got = lognorm.cdf(x, 1.5, 0.1, 1.2)
    expected = sc.lognorm.cdf(x, 1.5, 0.1, 1.2)
    np.testing.assert_allclose(got, expected)


def test_lognorm_ppf():
    from numba_stats import lognorm

    p = np.linspace(0, 1, 10)
    got = lognorm.ppf(p, 1.5, 0.1, 1.2)
    expected = sc.lognorm.ppf(p, 1.5, 0.1, 1.2)
    np.testing.assert_allclose(got, expected)



"""
Empirical density distribution formed by a Bernstein polynomial.

The Bernstein polynomial basis is better suited to model a probability distribution
than the Chebychev basis, since it is possible to implement the constraint
f(x; p) >= 0 with simple parameter limits p >= 0 (where p is a vector).

The density function and its integral are not normalised. This is not an issue when
the density is used in an extended maximum-likelihood fit.
"""

import numpy as np
import numba as nb


_signatures = [
    (nb.float32[:], nb.float32[:], nb.float32[:]),
    (nb.float64[:], nb.float64[:], nb.float64[:]),
]


@nb.njit(_signatures, cache=True)
def _de_castlejau(z, beta, res):
    # De Casteljau algorithm, numerically stable
    n = len(beta)
    if n == 0:
        res[:] = np.nan
    else:
        betai = np.empty_like(beta)
        for iz, zi in enumerate(z):
            azi = 1.0 - zi
            betai[:] = beta
            for j in range(1, n):
                for k in range(n - j):
                    betai[k] = betai[k] * azi + betai[k + 1] * zi
            res[iz] = betai[0]
    return res


_signatures = [
    nb.float32[:](nb.float32[:]),
    nb.float64[:](nb.float64[:]),
]


@nb.njit(_signatures, cache=True)
def _beta_int(beta):
    n = len(beta)
    r = np.zeros(n + 1, dtype=beta.dtype)
    for j in range(1, n + 1):
        for k in range(j):
            r[j] += beta[k]
    r *= 1.0 / n
    return r


@nb.njit(cache=True)
def _prepare_z_beta(x, xmin, xmax, beta):
    inverse_scale = 1 / (xmax - xmin)
    z = x.copy()
    z -= xmin
    z *= inverse_scale
    # beta = beta.copy()
    # inverse_scale /= len(beta) + 1
    # beta *= inverse_scale
    return z, beta


def _prepare_array(x):
    x = np.atleast_1d(x)
    if x.dtype.kind != "f":
        x = x.astype(np.float64)
    return x


_signatures = [
    (nb.float32[:], nb.float32[:], nb.float32[:], nb.float32[:], nb.float32[:]),
    (nb.float64[:], nb.float64[:], nb.float64[:], nb.float64[:], nb.float64[:]),
]


@nb.guvectorize(_signatures, "(),(n),(),()->()", cache=True)
def scaled_pdf(x, beta, xmin, xmax, res):
    z, beta = _prepare_z_beta(x, xmin, xmax, beta)
    _de_castlejau(z, beta, res)


@nb.guvectorize(_signatures, "(),(n),(),()->()", cache=True)
def scaled_cdf(x, beta, xmin, xmax, res):
    z, beta = _prepare_z_beta(x, xmin, xmax, beta)
    beta = _beta_int(beta)
    _de_castlejau(z, beta, res)


@nb.extending.overload(scaled_pdf)
def bernstein_scaled_pdf_ol(x, beta, xmin, xmax):
    from numba.core.errors import TypingError
    from numba.types import Array, Float

    if not isinstance(x, Array):
        raise TypingError("x must be a Numpy array")
    if not isinstance(beta, Array):
        raise TypingError("beta must be a Numpy array")
    if not isinstance(xmin, Float):
        raise TypingError("xmin must be float")
    if not isinstance(xmax, Float):
        raise TypingError("xmax must be float")

    def impl(x, beta, xmin, xmax):
        z, beta = _prepare_z_beta(x, xmin, xmax, beta)
        res = np.empty_like(z)
        _de_castlejau(z, beta, res)
        return res

    return impl


@nb.extending.overload(scaled_cdf)
def bernstein_scaled_cdf_ol(x, beta, xmin, xmax):
    from numba.core.errors import TypingError
    from numba.types import Array, Float

    if not isinstance(x, Array):
        raise TypingError("x must be a Numpy array")
    if not isinstance(beta, Array):
        raise TypingError("beta must be a Numpy array")
    if not isinstance(xmin, Float):
        raise TypingError("xmin must be float")
    if not isinstance(xmax, Float):
        raise TypingError("xmax must be float")

    def impl(x, beta, xmin, xmax):
        z, beta = _prepare_z_beta(x, xmin, xmax, beta)
        beta = _beta_int(beta)
        res = np.empty_like(z)
        _de_castlejau(z, beta, res)
        return res

    return impl


density = scaled_pdf



import numpy as np
from numba_stats import poisson
import scipy.stats as sc


def test_poisson_pmf():
    m = np.linspace(0, 3, 20)[:, np.newaxis]
    k = np.arange(10)
    got = poisson.pmf(k, m)
    expected = sc.poisson.pmf(k, m)
    np.testing.assert_allclose(got, expected)


def test_poisson_cdf():
    m = np.linspace(0, 3, 20)[:, np.newaxis]
    k = np.arange(10)
    got = poisson.cdf(k, m)
    expected = sc.poisson.cdf(k, m)
    np.testing.assert_allclose(got, expected)



"""
Lognormal distribution.
"""
import numpy as np
from .norm import _cdf, _ppf
from ._util import _jit, _vectorize


# has to be separate to avoid a warning
@_jit
def _logpdf(x, s, loc, scale):
    z = (x - loc) / scale
    if z <= 0:
        return -np.inf
    c = np.sqrt(2 * np.pi)
    log_pdf = -0.5 * np.log(z) ** 2 / s ** 2 - np.log(s * z * c)
    return log_pdf - np.log(scale)


@_vectorize(4)
def logpdf(x, s, loc, scale):
    """
    Return log of probability density.
    """
    return _logpdf(x, s, loc, scale)


@_vectorize(4)
def pdf(x, s, loc, scale):
    """
    Return probability density.
    """
    return np.exp(_logpdf(x, s, loc, scale))


@_vectorize(4)
def cdf(x, s, loc, scale):
    """
    Return cumulative probability.
    """
    z = (x - loc) / scale
    if z <= 0:
        return 0.0
    return _cdf(np.log(z) / s)


@_vectorize(4, cache=False)  # no cache because of _ppf
def ppf(p, s, loc, scale):
    """
    Return quantile for given probability.
    """
    z = np.exp(s * _ppf(p))
    return scale * z + loc



"""
Tsallis-Hagedorn distribution.

A generalisation (q-analog) of the exponential distribution based on Tsallis entropy. It
approximately describes the pT distribution charged particles produced in high-energy
minimum bias particle collisions.
"""

import numpy as np
from ._util import _vectorize


@_vectorize(4)
def pdf(x, m, t, n):
    """
    Return probability density.
    """
    # Formula from CMS, Eur. Phys. J. C (2012) 72:2164
    if n <= 2:
        raise ValueError("n > 2 is required")

    mt = np.sqrt(m ** 2 + x ** 2)
    nt = n * t
    c = (n - 1) * (n - 2) / (nt * (nt + (n - 2) * m))
    return c * x * (1 + (mt - m) / nt) ** -n


@_vectorize(4)
def cdf(x, m, t, n):
    """
    Return cumulative probability.
    """
    # Formula computed from tsallis_pdf with Sympy, then simplified by hand
    if n <= 2:
        raise ValueError("n > 2 is required")

    mt = np.sqrt(m ** 2 + x ** 2)
    nt = n * t
    return ((mt - m) / nt + 1) ** (1 - n) * (m + mt - n * (mt + t)) / (m * (n - 2) + nt)



"""
Uniform distribution.
"""
from ._util import _vectorize
import numpy as np


@_vectorize(3)
def logpdf(x, a, w):
    """
    Return probability density.
    """
    if a <= x <= a + w:
        return -np.log(w)
    return -np.inf


@_vectorize(3)
def pdf(x, a, w):
    """
    Return probability density.
    """
    return np.exp(logpdf(x, a, w))


@_vectorize(3)
def cdf(x, a, w):
    """
    Return cumulative probability.
    """
    if a <= x:
        if x <= a + w:
            return (x - a) / w
        return 1
    return 0


@_vectorize(3)
def ppf(p, a, w):
    """
    Return quantile for given probability.
    """
    return w * p + a



"""
Voigtian distribution.

This is the convolution of a Cauchy distribution with a normal distribution.

There is a closed form for the cdf, but the required hypergeometric function is not
implemented anywhere.

https://en.wikipedia.org/wiki/Voigt_profile
"""

from ._special import voigt_profile as _voigt
from ._util import _vectorize


@_vectorize(4, cache=False)
def pdf(x, gamma, loc, scale):
    """
    Return probability density.
    """
    return _voigt(x - loc, scale, gamma)



"""
Student's t distribution.
"""
import numpy as np
from ._special import stdtr as _cdf, stdtrit as _ppf
from ._util import _vectorize
from math import lgamma as _lgamma


@_vectorize(4, cache=False)
def logpdf(x, df, loc, scale):
    """
    Return probability density.
    """
    z = (x - loc) / scale
    k = 0.5 * (df + 1)
    logp = _lgamma(k) - _lgamma(0.5 * df)
    logp -= 0.5 * np.log(df * np.pi) + k * np.log(1 + (z ** 2) / df) + np.log(scale)
    return logp


@_vectorize(4, cache=False)
def pdf(x, df, loc, scale):
    """
    Return probability density.
    """
    return np.exp(logpdf(x, df, loc, scale))


@_vectorize(4, cache=False)
def cdf(x, df, loc, scale):
    """
    Return cumulative probability.
    """
    z = (x - loc) / scale
    return _cdf(df, z)


@_vectorize(4, cache=False)
def ppf(p, df, loc, scale):
    """
    Return quantile for given probability.
    """
    if p == 0:
        return -np.inf
    elif p == 1:
        return np.inf
    z = _ppf(df, p)
    return scale * z + loc



# numba currently does not support scipy, so we cannot access
# scipy.stats.norm.ppf and scipy.stats.poisson.cdf in a JIT'ed
# function. As a workaround, we wrap special functions from
# scipy to implement the needed functions here.
from numba.extending import get_cython_function_address
from numba.types import WrapperAddressProtocol, float64
import scipy.special.cython_special as cysp


def get(name, signature):
    # create new function object with correct signature that numba can call by extracting
    # function pointer from scipy.special.cython_special; uses scipy/cython internals
    index = 1 if signature.return_type is float64 else 0
    pyx_fuse_name = f"__pyx_fuse_{index}{name}"
    if pyx_fuse_name in cysp.__pyx_capi__:
        name = pyx_fuse_name
    addr = get_cython_function_address("scipy.special.cython_special", name)

    # dynamically create type that inherits from WrapperAddressProtocol
    cls = type(
        name,
        (WrapperAddressProtocol,),
        {"__wrapper_address__": lambda self: addr, "signature": lambda self: signature},
    )
    return cls()


# unary functions (double)
erfinv = get("erfinv", float64(float64))

# binary functions (double)
gammaincc = get("gammaincc", float64(float64, float64))
stdtrit = get("stdtrit", float64(float64, float64))
stdtr = get("stdtr", float64(float64, float64))
betainc = get("betainc", float64(float64, float64, float64))

# n-ary functions (double)
voigt_profile = get("voigt_profile", float64(float64, float64, float64))



"""
Truncated exponential distribution.
"""
import numpy as np
from ._util import _vectorize
from .expon import _cdf, _ppf


@_vectorize(5)
def logpdf(x, xmin, xmax, loc, scale):
    """
    Return log of probability density.
    """
    if x < xmin:
        return -np.inf
    elif x > xmax:
        return -np.inf
    scale_inv = 1 / scale
    z = (x - loc) * scale_inv
    zmin = (xmin - loc) * scale_inv
    zmax = (xmax - loc) * scale_inv
    return -z + np.log(scale_inv / (_cdf(zmax) - _cdf(zmin)))


@_vectorize(5)
def pdf(x, xmin, xmax, loc, scale):
    """
    Return probability density.
    """
    return np.exp(logpdf(x, xmin, xmax, loc, scale))


@_vectorize(5)
def cdf(x, xmin, xmax, loc, scale):
    """
    Return cumulative probability.
    """
    if x < xmin:
        return 0.0
    elif x > xmax:
        return 1.0
    scale_inv = 1 / scale
    z = (x - loc) * scale_inv
    p = _cdf(z)
    zmin = (xmin - loc) * scale_inv
    zmax = (xmax - loc) * scale_inv
    pmin = _cdf(zmin)
    pmax = _cdf(zmax)
    return (p - pmin) / (pmax - pmin)


@_vectorize(5)
def ppf(p, xmin, xmax, loc, scale):
    """
    Return quantile for given probability.
    """
    scale_inv = 1 / scale
    zmin = (xmin - loc) * scale_inv
    zmax = (xmax - loc) * scale_inv
    pmin = _cdf(zmin)
    pmax = _cdf(zmax)
    pstar = p * (pmax - pmin) + pmin
    z = _ppf(pstar)
    x = z * scale + loc
    return x



import scipy.stats as sc
import numpy as np


def test_norm_pdf():
    from numba_stats import norm

    x = np.linspace(-5, 5, 10)
    got = norm.pdf(x, 1, 2)
    expected = sc.norm.pdf(x, 1, 2)
    np.testing.assert_allclose(got, expected)


def test_norm_logpdf():
    from numba_stats import norm

    x = np.linspace(-5, 5, 10)
    got = norm.logpdf(x, 1, 2)
    expected = sc.norm.logpdf(x, 1, 2)
    np.testing.assert_allclose(got, expected)


def test_norm_cdf():
    from numba_stats import norm

    x = np.linspace(-5, 5, 10)
    got = norm.cdf(x, 1, 2)
    expected = sc.norm.cdf(x, 1, 2)
    np.testing.assert_allclose(got, expected)


def test_norm_ppf():
    from numba_stats import norm

    p = np.linspace(0, 1, 10)
    got = norm.ppf(p, 0, 1)
    expected = sc.norm.ppf(p)
    np.testing.assert_allclose(got, expected)



from scipy.special import voigt_profile
from numba_stats import voigt
import numpy as np


def test_voigt_pdf():
    gamma = 2
    sigma = 1
    mu = -1
    x = np.linspace(-5, 5, 10)
    got = voigt.pdf(x, gamma, mu, sigma)
    # note that sigma comes before gamma in scipy
    expected = voigt_profile(x - mu, sigma, gamma)
    np.testing.assert_allclose(got, expected)



import numpy as np
from numba_stats import truncexpon, expon


def test_pdf():
    x = np.linspace(-1, 5, 10)
    xmin = 1
    xmax = 4
    mu = 2
    sigma = 3
    got = truncexpon.pdf(x, xmin, xmax, mu, sigma)
    expected = expon.pdf(x, mu, sigma) / (
        expon.cdf(xmax, mu, sigma) - expon.cdf(xmin, mu, sigma)
    )
    expected[x < xmin] = 0
    expected[x > xmax] = 0
    np.testing.assert_allclose(got, expected)


def test_cdf():
    x = np.linspace(-1, 5, 10)
    xmin = 1
    xmax = 4
    mu = 1
    sigma = 3
    got = truncexpon.cdf(x, xmin, xmax, mu, sigma)
    expected = (expon.cdf(x, mu, sigma) - expon.cdf(xmin, mu, sigma)) / (
        expon.cdf(xmax, mu, sigma) - expon.cdf(xmin, mu, sigma)
    )
    expected[x < xmin] = 0
    expected[x > xmax] = 1
    np.testing.assert_allclose(got, expected)


def test_ppf():
    expected = np.linspace(0, 1, 10)
    xmin = 1
    xmax = 4
    mu = 2
    sigma = 3
    x = truncexpon.ppf(expected, xmin, xmax, mu, sigma)
    got = truncexpon.cdf(x, xmin, xmax, mu, sigma)
    np.testing.assert_allclose(got, expected, atol=1e-14)
